#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from enum import Enum

from octobot_channels.channels.channel_instances import get_chan_at_id
from octobot_commons.channels_name import OctoBotChannelsName
from octobot_commons.logging.logging_util import get_logger

from octobot_commons.enums import OctoBotChannelSubjects

from octobot_evaluators.api.evaluators import create_all_type_evaluators

OCTOBOT_CHANNEL_EVALUATOR_CONSUMER_LOGGER_TAG = "OctoBotChannelEvaluatorConsumer"


class OctoBotChannelEvaluatorActions(Enum):
    """
    OctoBot Channel consumer supported actions
    """

    EVALUATOR = "evaluator"


class OctoBotChannelEvaluatorDataKeys(Enum):
    """
    OctoBot Channel consumer supported data keys
    """

    EXCHANGE_CONFIGURATION = "exchange_configuration"
    MATRIX_ID = "matrix_id"
    TENTACLES_SETUP_CONFIG = "tentacles_setup_config"


async def octobot_channel_callback(bot_id, subject, action, data) -> None:
    """
    OctoBot channel consumer callback
    :param bot_id: the callback bot id
    :param subject: the callback subject
    :param action: the callback action
    :param data: the callback data
    """
    if subject == OctoBotChannelSubjects.CREATION.value:
        await _handle_creation(bot_id, action, data)


async def _handle_creation(bot_id, action, data):
    if action == OctoBotChannelEvaluatorActions.EVALUATOR.value:
        try:
            exchange_configuration = data[OctoBotChannelEvaluatorDataKeys.EXCHANGE_CONFIGURATION.value]
            await create_all_type_evaluators(
                tentacles_setup_config=data[OctoBotChannelEvaluatorDataKeys.TENTACLES_SETUP_CONFIG.value],
                matrix_id=data[OctoBotChannelEvaluatorDataKeys.MATRIX_ID.value],
                exchange_name=exchange_configuration.exchange_name,
                bot_id=bot_id,
                symbols_by_crypto_currencies=exchange_configuration.symbols_by_crypto_currencies,
                symbols=exchange_configuration.symbols,
                time_frames=exchange_configuration.time_frames_without_real_time,
                real_time_time_frames=exchange_configuration.real_time_time_frames
            )
            await get_chan_at_id(OctoBotChannelsName.OCTOBOT_CHANNEL.value, bot_id).get_internal_producer() \
                .send(bot_id=bot_id,
                      subject=OctoBotChannelSubjects.NOTIFICATION.value,
                      action=action,
                      data={OctoBotChannelEvaluatorDataKeys.MATRIX_ID.value:
                            data[OctoBotChannelEvaluatorDataKeys.MATRIX_ID.value]})

        except Exception as e:
            get_logger(OCTOBOT_CHANNEL_EVALUATOR_CONSUMER_LOGGER_TAG).error(f"Error when creating new evaluator {e}")



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
from mock import patch
import octobot_evaluators.api.evaluators
from octobot_commons.enums import TimeFrames
from octobot_commons.symbol_util import split_symbol


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def symbols_by_crypto_currencies():
    return {"Bitcoin": ["BTC/USD", "BTC/USDT"], "Ethereum": ["ETH/BTC"]}


@pytest.fixture
def symbols():
    return ["BTC/USD", "BTC/USDT", "ETH/BTC"]


@pytest.fixture
def time_frames():
    return [TimeFrames.THREE_DAYS, TimeFrames.ONE_HOUR]


async def test_create_evaluators_no_wild_card(symbols_by_crypto_currencies, symbols, time_frames):
    evaluators = await _create_evaluators(EvaluatorNoWildCard, symbols_by_crypto_currencies, symbols, time_frames)
    assert len(evaluators) == 6

    # one eval per time frame per symbol (1 symbols)
    assert len([e for e in evaluators if e.cryptocurrency == "ETH"]) == len(time_frames) * 1
    # one eval per time frame per symbol (2 symbols)
    assert len([e for e in evaluators if e.cryptocurrency == "BTC"]) == len(time_frames) * 2

    # right crypto-currency name
    assert all([e.cryptocurrency_name == "Ethereum" for e in evaluators if e.cryptocurrency == "ETH"])
    assert all([e.cryptocurrency_name == "Bitcoin" for e in evaluators if e.cryptocurrency == "BTC"])

    # right crypto-currency
    assert all([e.cryptocurrency == "ETH" for e in evaluators if e.symbol == "ETH/BTC"])
    assert all([e.cryptocurrency == "BTC" for e in evaluators if e.symbol in ["BTC/USD", "BTC/USDT"]])

    # right symbol
    assert all([e.symbol == "ETH/BTC" for e in evaluators if e.cryptocurrency == "ETH"])
    assert all([e.symbol in ["BTC/USD", "BTC/USDT"] for e in evaluators if e.cryptocurrency == "BTC"])
    # all symbols taken into account
    assert len(set([e.symbol for e in evaluators if e.cryptocurrency == "BTC"])) == 2
    assert len(set([e.symbol for e in evaluators if e.cryptocurrency == "ETH"])) == 1

    # valid time frames
    assert all([e.time_frame in time_frames for e in evaluators])
    # all time frames taken into account
    assert len(set([e.time_frame for e in evaluators if e.symbol == "BTC/USD"])) == 2
    assert len(set([e.time_frame for e in evaluators if e.symbol == "BTC/USDT"])) == 2
    assert len(set([e.time_frame for e in evaluators if e.symbol == "ETH/BTC"])) == 2


async def test_create_evaluators_wild_card(symbols_by_crypto_currencies, symbols, time_frames):
    evaluators = await _create_evaluators(EvaluatorWildCard, symbols_by_crypto_currencies, symbols, time_frames)

    # only one wild card evaluator created
    assert len(evaluators) == 1
    assert evaluators[0].cryptocurrency is evaluators[0].cryptocurrency_name is evaluators[0].symbol is \
        evaluators[0].time_frame is None


async def test_create_evaluators_no_cc_wild_card(symbols_by_crypto_currencies, symbols, time_frames):
    evaluators = await _create_evaluators(EvaluatorNoCCWildCard, symbols_by_crypto_currencies, symbols, time_frames)

    # only difference is the cryptocurrency attribute
    assert len(evaluators) == 2
    assert evaluators[0].cryptocurrency_name is evaluators[0].symbol is evaluators[0].time_frame is None
    assert sorted([e.cryptocurrency for e in evaluators]) == sorted(["BTC", "ETH"])


async def test_create_evaluators_no_cc_name_wild_card(symbols_by_crypto_currencies, symbols, time_frames):
    evaluators = await _create_evaluators(EvaluatorNoCCNameWildCard, symbols_by_crypto_currencies, symbols, time_frames)

    # only one wild card evaluator created since EvaluatorNoCCNameWildCard is crypto-currency wild card
    assert len(evaluators) == 1
    assert evaluators[0].cryptocurrency is evaluators[0].cryptocurrency_name is evaluators[0].symbol is \
        evaluators[0].time_frame is None


async def test_create_evaluators_no_cc_name_and_no_cc_wild_card(symbols_by_crypto_currencies, symbols, time_frames):
    evaluators = await _create_evaluators(EvaluatorNoCCNameNoCCWildCard, symbols_by_crypto_currencies,
                                          symbols, time_frames)

    # only differences are the cryptocurrency and cryptocurrency_name attribute
    assert len(evaluators) == 2
    assert evaluators[0].symbol is evaluators[0].time_frame is None
    assert sorted([e.cryptocurrency for e in evaluators]) == sorted(["BTC", "ETH"])
    assert sorted([e.cryptocurrency_name for e in evaluators]) == sorted(["Bitcoin", "Ethereum"])


async def test_create_evaluators_no_symbol_wild_card(symbols_by_crypto_currencies, symbols, time_frames):
    evaluators = await _create_evaluators(EvaluatorNoSymbolWildCard, symbols_by_crypto_currencies, symbols, time_frames)

    # only difference is the symbol attribute
    assert len(evaluators) == 3
    assert evaluators[0].cryptocurrency is evaluators[0].cryptocurrency_name is evaluators[0].time_frame is None
    assert [e.symbol for e in evaluators] == symbols


async def test_create_evaluators_no_time_frame_wild_card(symbols_by_crypto_currencies, symbols, time_frames):
    evaluators = await _create_evaluators(EvaluatorNoTimeFrameWildCard, symbols_by_crypto_currencies,
                                          symbols, time_frames)

    # only difference is the time_frame attribute
    assert len(evaluators) == 2
    assert evaluators[0].cryptocurrency is evaluators[0].cryptocurrency_name is evaluators[0].symbol is None
    assert [e.time_frame for e in evaluators] == time_frames


async def test_create_evaluators_no_cc_name_no_symbol_wild_card(symbols_by_crypto_currencies, symbols, time_frames):
    evaluators = await _create_evaluators(EvaluatorNoCCNameSymbolWildCard, symbols_by_crypto_currencies,
                                          symbols, time_frames)

    # only difference is the symbol attribute since no crypto-currency wildcard name requires also no
    # cryptocurrency wildcard
    assert len(evaluators) == 3
    assert evaluators[0].cryptocurrency is evaluators[0].cryptocurrency_name is evaluators[0].time_frame is None
    assert [e.symbol for e in evaluators] == symbols


async def test_create_evaluators_no_cc_time_frame_wild_card(symbols_by_crypto_currencies, symbols, time_frames):
    evaluators = await _create_evaluators(EvaluatorNoCCTimeFrameWildCard, symbols_by_crypto_currencies,
                                          symbols, time_frames)

    # only differences are on cryptocurrency and time frame attributes
    assert len(evaluators) == 4
    assert evaluators[0].symbol is evaluators[0].cryptocurrency_name is None
    assert sorted(list(set(e.cryptocurrency for e in evaluators))) == sorted(["ETH", "BTC"])
    tfs = set(e.time_frame for e in evaluators)
    assert len(tfs) == 2
    assert all(tf in time_frames for tf in tfs)


async def test_create_evaluators_no_symbol_time_frame_wild_card(symbols_by_crypto_currencies, symbols, time_frames):
    evaluators = await _create_evaluators(EvaluatorNoSymbolTimeFrameWildCard, symbols_by_crypto_currencies,
                                          symbols, time_frames)

    # only differences are on symbol and time frame attributes
    assert len(evaluators) == 6
    assert evaluators[0].cryptocurrency is evaluators[0].cryptocurrency_name is None
    assert sorted(list(set(e.symbol for e in evaluators))) == sorted(symbols)
    tfs = set(e.time_frame for e in evaluators)
    assert len(tfs) == 2
    assert all(tf in time_frames for tf in tfs)


async def _create_evaluators(evaluator_parent_class, symbols_by_crypto_currencies, symbols, time_frames):
    crypto_currency_name_by_crypto_currencies = {}
    symbols_by_crypto_currency_tickers = {}
    for name, symbol_list in symbols_by_crypto_currencies.items():
        ticker = split_symbol(symbol_list[0])[0]
        crypto_currency_name_by_crypto_currencies[ticker] = name
        symbols_by_crypto_currency_tickers[ticker] = symbol_list
    with patch("octobot_evaluators.api.evaluators.create_evaluator", new=_mocked_create_evaluator), \
            patch("octobot_evaluators.api.evaluators.get_all_classes_from_parent",
                  new=_mocked_get_all_classes_from_parent):
        return await octobot_evaluators.api.evaluators.create_evaluators(
            evaluator_parent_class=evaluator_parent_class,
            tentacles_setup_config=None,
            matrix_id="",
            exchange_name="",
            bot_id="",
            crypto_currency_name_by_crypto_currencies=crypto_currency_name_by_crypto_currencies,
            symbols_by_crypto_currency_tickers=symbols_by_crypto_currency_tickers,
            symbols=symbols,
            time_frames=time_frames
        )


def _mocked_get_all_classes_from_parent(evaluator_parent_class):
    return [evaluator_parent_class]


async def _mocked_create_evaluator(evaluator_class,
                                   tentacles_setup_config: object,
                                   matrix_id: str,
                                   exchange_name: str,
                                   bot_id: str,
                                   cryptocurrency: str = None,
                                   cryptocurrency_name: str = None,
                                   symbol: str = None,
                                   time_frame=None,
                                   relevant_evaluators=None,
                                   all_symbols_by_crypto_currencies=None,
                                   time_frames=None,
                                   real_time_time_frames=None):
    return evaluator_class(cryptocurrency, cryptocurrency_name, symbol, time_frame, all_symbols_by_crypto_currencies)


class EvaluatorWildCard:
    def __init__(self, cryptocurrency, cryptocurrency_name, symbol, time_frame, all_symbols_by_crypto_currencies):
        self.cryptocurrency = cryptocurrency
        self.cryptocurrency_name = cryptocurrency_name
        self.symbol = symbol
        self.time_frame = time_frame
        self.all_symbols_by_crypto_currencies = all_symbols_by_crypto_currencies

    @classmethod
    def get_is_cryptocurrency_name_wildcard(cls):
        return True

    @classmethod
    def get_is_symbol_wildcard(cls):
        return True

    @classmethod
    def get_is_time_frame_wildcard(cls):
        return True


    @classmethod
    def get_is_cryptocurrencies_wildcard(cls):
        return True


class EvaluatorNoCCWildCard(EvaluatorWildCard):
    @classmethod
    def get_is_cryptocurrencies_wildcard(cls):
        return False


class EvaluatorNoCCNameWildCard(EvaluatorWildCard):
    @classmethod
    def get_is_cryptocurrency_name_wildcard(cls):
        return False


class EvaluatorNoSymbolWildCard(EvaluatorWildCard):
    @classmethod
    def get_is_symbol_wildcard(cls):
        return False


class EvaluatorNoTimeFrameWildCard(EvaluatorWildCard):
    @classmethod
    def get_is_time_frame_wildcard(cls):
        return False


class EvaluatorNoCCNameNoCCWildCard(EvaluatorNoCCNameWildCard, EvaluatorNoCCWildCard):
    pass


class EvaluatorNoCCNameSymbolWildCard(EvaluatorNoCCNameWildCard, EvaluatorNoSymbolWildCard):
    pass


class EvaluatorNoCCTimeFrameWildCard(EvaluatorNoCCWildCard, EvaluatorNoTimeFrameWildCard):
    pass


class EvaluatorNoSymbolTimeFrameWildCard(EvaluatorNoSymbolWildCard, EvaluatorNoTimeFrameWildCard):
    pass


class EvaluatorNoWildCard(EvaluatorNoCCWildCard, EvaluatorNoCCNameWildCard,
                          EvaluatorNoSymbolWildCard, EvaluatorNoTimeFrameWildCard):
    def __init__(self, *args):
        super().__init__(*args)



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_channels.util.channel_creator import create_all_subclasses_channel
from octobot_commons.constants import CONFIG_TIME_FRAME
from octobot_commons.tentacles_management.class_inspector import get_all_classes_from_parent
from octobot_commons.time_frame_manager import sort_time_frames
from octobot_evaluators.channels.evaluator_channel import get_chan, set_chan, EvaluatorChannel, del_chan
from octobot_evaluators.constants import MATRIX_CHANNEL
from octobot_evaluators.evaluator import StrategyEvaluator, EVALUATORS_CHANNEL
from octobot_tentacles_manager.api.configurator import is_tentacle_activated_in_tentacles_setup_config


def init_time_frames_from_strategies(config, tentacles_setup_config) -> None:
    time_frame_list = set()
    for strategies_eval_class in get_activated_strategies_classes(tentacles_setup_config):
        for time_frame in strategies_eval_class.get_required_time_frames(config):
            time_frame_list.add(time_frame)
    time_frame_list = sort_time_frames(list(time_frame_list))
    config[CONFIG_TIME_FRAME] = time_frame_list


def get_activated_strategies_classes(tentacles_setup_config):
    return [
        strategies_eval_class
        for strategies_eval_class in get_all_classes_from_parent(StrategyEvaluator)
        if is_tentacle_activated_in_tentacles_setup_config(tentacles_setup_config, strategies_eval_class.get_name())
    ]


async def create_evaluator_channels(matrix_id: str, is_backtesting: bool = False) -> None:
    await create_all_subclasses_channel(EvaluatorChannel, set_chan, is_synchronized=is_backtesting, matrix_id=matrix_id)


def del_evaluator_channels(matrix_id: str) -> None:
    del_chan(MATRIX_CHANNEL, matrix_id)
    del_chan(EVALUATORS_CHANNEL, matrix_id)


def matrix_channel_exists(matrix_id: str) -> bool:
    try:
        get_chan(MATRIX_CHANNEL, matrix_id)
        return True
    except KeyError:
        return False



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from typing import NewType, Union

from octobot_evaluators.enums import EvaluatorMatrixTypes

MatrixValueType = NewType('MatrixValueType', Union[str, int, float])

START_EVAL_PERTINENCE = 1
MAX_TA_EVAL_TIME_SECONDS = 0.1
EVALUATOR_EVAL_DEFAULT_TYPE = float
STRATEGIES_REQUIRED_TIME_FRAME = "required_time_frames"
TRADING_MODE_REQUIRED_STRATEGIES = "required_strategies"
TRADING_MODE_REQUIRED_STRATEGIES_MIN_COUNT = "required_strategies_min_count"
STRATEGIES_REQUIRED_EVALUATORS = "required_evaluators"
STRATEGIES_COMPATIBLE_EVALUATOR_TYPES = "compatible_evaluator_types"
CONFIG_FORCED_TIME_FRAME = "forced_time_frame"
TENTACLE_DEFAULT_CONFIG = "default_config"

EVALUATOR_CLASS_TYPE_MRO_INDEX = -4

EVALUATORS_CHANNEL: str = "Evaluators"
MATRIX_CHANNEL: str = "Matrix"
MATRIX_CHANNELS: str = "MatrixChannels"

TA_RE_EVALUATION_TRIGGER_UPDATED_DATA = "TA_re_evaluation_trigger_updated_data"
RESET_EVALUATION = "reset_evaluation"
EVALUATOR_CHANNEL_DATA_ACTION = "action"
EVALUATOR_CHANNEL_DATA_EXCHANGE_ID = "exchange_id"
EVALUATOR_CHANNEL_DATA_TIME_FRAMES = "time_frames"

evaluator_class_str_to_matrix_type_dict = {
    "TAEvaluator": EvaluatorMatrixTypes.TA,
    "SocialEvaluator": EvaluatorMatrixTypes.SOCIAL,
    "RealTimeEvaluator": EvaluatorMatrixTypes.REAL_TIME,
    "StrategyEvaluator": EvaluatorMatrixTypes.STRATEGIES
}



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_evaluators.api.evaluators import create_matrix
from octobot_evaluators.api.initialization import create_evaluator_channels, del_evaluator_channels
from octobot_evaluators.channels.evaluator_channel import del_chan, get_chan
from octobot_evaluators.constants import MATRIX_CHANNEL
from octobot_evaluators.data_manager.matrix_manager import get_tentacle_path
from octobot_evaluators.matrices.matrices import Matrices

MATRIX_TEST_ID = "test"


async def matrix_callback(matrix_id,
                          evaluator_name,
                          evaluator_type,
                          eval_note,
                          eval_note_type,
                          eval_time,
                          exchange_name,
                          cryptocurrency,
                          symbol,
                          time_frame):
    pass


@pytest.mark.asyncio
async def test_evaluator_channel_creation():
    del_evaluator_channels(MATRIX_TEST_ID)
    await create_evaluator_channels(MATRIX_TEST_ID)
    await get_chan(MATRIX_CHANNEL, MATRIX_TEST_ID).new_consumer(matrix_callback)
    await get_chan(MATRIX_CHANNEL, MATRIX_TEST_ID).stop()


@pytest.mark.asyncio
async def test_evaluator_channel_send():
    del_evaluator_channels(MATRIX_TEST_ID)
    matrix_id = create_matrix()
    await create_evaluator_channels(MATRIX_TEST_ID)
    await get_chan(MATRIX_CHANNEL, MATRIX_TEST_ID).new_consumer(matrix_callback)
    await get_chan(MATRIX_CHANNEL, MATRIX_TEST_ID).get_internal_producer().send(matrix_id=matrix_id,
                                                                                evaluator_name="test",
                                                                                evaluator_type="test2",
                                                                                eval_note=1)

    # following assert should be None because send() doesn't call set_tentacle_value
    assert Matrices.instance().get_matrix(matrix_id).get_node_at_path(
        get_tentacle_path(tentacle_name="test", tentacle_type="test2")) is None
    await get_chan(MATRIX_CHANNEL, MATRIX_TEST_ID).stop()
    Matrices.instance().del_matrix(matrix_id)


@pytest.mark.asyncio
async def test_evaluator_channel_send_eval_note():
    del_evaluator_channels(MATRIX_TEST_ID)
    matrix_id = create_matrix()
    await create_evaluator_channels(MATRIX_TEST_ID)
    await get_chan(MATRIX_CHANNEL, MATRIX_TEST_ID).new_consumer(matrix_callback)
    await get_chan(MATRIX_CHANNEL, MATRIX_TEST_ID).get_internal_producer().send_eval_note(matrix_id=matrix_id,
                                                                                          evaluator_name="test",
                                                                                          evaluator_type="test2",
                                                                                          eval_note=1,
                                                                                          eval_note_type=int)

    assert Matrices.instance().get_matrix(matrix_id).get_node_at_path(
        get_tentacle_path(tentacle_name="test", tentacle_type="test2")).node_value == 1
    await get_chan(MATRIX_CHANNEL, MATRIX_TEST_ID).stop()
    Matrices.instance().del_matrix(matrix_id)



# cython: language_level=3
#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_commons.event_tree cimport EventTreeNode
from octobot_evaluators.data.matrix cimport Matrix

cpdef Matrix get_matrix(str matrix_id)
cpdef void set_tentacle_value(str matrix_id, list tentacle_path, object tentacle_type,
                              object tentacle_value, double timestamp=*)
cpdef EventTreeNode get_tentacle_node(str matrix_id, list tentacle_path)
cpdef object get_tentacle_value(str matrix_id, list tentacle_path)
cpdef list get_matrix_default_value_path(str tentacle_name,
                                         object tentacle_type,
                                         str exchange_name=*,
                                         str cryptocurrency=*,
                                         str symbol=*,
                                         str time_frame=*)
cpdef object get_tentacle_eval_time(str matrix_id, list tentacle_path)
cpdef list get_tentacle_nodes(str matrix_id, str exchange_name=*, object tentacle_type=*, str tentacle_name=*)
cpdef dict get_node_children_by_names_at_path(str matrix_id, list tentacle_path, object starting_node=*)

cpdef list get_tentacles_value_nodes(str matrix_id,
                                     list tentacle_nodes,
                                     str cryptocurrency=*,
                                     str symbol=*,
                                     str time_frame=*)
cpdef list get_tentacle_path(str exchange_name=*, object tentacle_type=*, str tentacle_name=*)
cpdef list get_tentacle_value_path(str cryptocurrency=*, str symbol=*, str time_frame=*)
cpdef dict get_evaluations_by_evaluator(str matrix_id,
                                        str exchange_name=*,
                                        str tentacle_type=*,
                                        str cryptocurrency=*,
                                        str symbol=*,
                                        str time_frame=*,
                                        bint allow_missing=*,
                                        list allowed_values=*)
cpdef list get_available_time_frames(str matrix_id,
                                     str exchange_name,
                                     str tentacle_type,
                                     str cryptocurrency,
                                     str symbol)
cpdef list get_available_symbols(str matrix_id,
                                 str exchange_name,
                                 str cryptocurrency,
                                 str tentacle_type=*,
                                 str second_tentacle_type=*)
cpdef bint is_tentacle_value_valid(str matrix_id, list tentacle_path, double timestamp=*, int delta=*)
cpdef bint is_tentacles_values_valid(str matrix_id, list tentacle_path_list, double timestamp=*, int delta=*)



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_evaluators.data.matrix import Matrix


def test_default_matrix():
    matrix = Matrix()
    assert matrix.matrix.root.children == {}


@pytest.mark.asyncio
async def test_get_node_at_path():
    matrix = Matrix()
    test_node_path = ["test-path", "test-path-2", "test-path3", 4]
    created_node = matrix.matrix.get_or_create_node(test_node_path)
    assert matrix.get_node_at_path(test_node_path) is created_node


@pytest.mark.asyncio
async def test_set_tentacle_value():
    matrix = Matrix()
    test_node_path = ["test-path", "test-path-2"]
    matrix.set_node_value("test-value", str, test_node_path)
    assert matrix.matrix.get_or_create_node(test_node_path).node_type == str
    assert matrix.matrix.get_or_create_node(test_node_path).node_value == "test-value"


@pytest.mark.asyncio
async def test_get_node_children_at_path():
    matrix = Matrix()
    test_node_1_path = ["test-path-parent", "test-path-child", "test-path-1"]
    test_node_2_path = ["test-path-parent", "test-path-child", "test-path-2"]
    test_node_3_path = ["test-path-parent", "test-path-child", "test-path-3"]
    created_node_1 = matrix.matrix.get_or_create_node(test_node_1_path)
    created_node_2 = matrix.matrix.get_or_create_node(test_node_2_path)
    created_node_3 = matrix.matrix.get_or_create_node(test_node_3_path)
    assert matrix.get_node_children_at_path(["test-path-parent", "test-path-child"]) == [created_node_1,
                                                                                         created_node_2,
                                                                                         created_node_3]


@pytest.mark.asyncio
async def test_get_node_children_by_names_at_path():
    matrix = Matrix()
    test_node_1_path = ["test-path-parent", "test-path-child", "test-path-1"]
    test_node_2_path = ["test-path-parent", "test-path-child", "test-path-2"]
    test_node_3_path = ["test-path-parent", "test-path-child", "test-path-3"]
    test_node_4_path = ["test-path-parent", "test-path-4"]
    test_node_5_path = ["test-path-parent", "test-path-4", "test-path-5"]
    test_node_6_path = ["test-path-parent", "test-path-child", "test-path-2", "test-path-6"]
    created_node_1 = matrix.matrix.get_or_create_node(test_node_1_path)
    created_node_2 = matrix.matrix.get_or_create_node(test_node_2_path)
    created_node_3 = matrix.matrix.get_or_create_node(test_node_3_path)
    created_node_4 = matrix.matrix.get_or_create_node(test_node_4_path)
    created_node_5 = matrix.matrix.get_or_create_node(test_node_5_path)
    created_node_6 = matrix.matrix.get_or_create_node(test_node_6_path)
    assert matrix.get_node_children_by_names_at_path(["test-path-parent", "test-path-child"]) == {
        "test-path-1": created_node_1,
        "test-path-2": created_node_2,
        "test-path-3": created_node_3
    }



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import time

import pytest
from octobot_commons.constants import MINUTE_TO_SECONDS, START_PENDING_EVAL_NOTE

from octobot_commons.enums import TimeFramesMinutes, TimeFrames

from octobot_evaluators.data.matrix import Matrix
from octobot_evaluators.data_manager.matrix_manager import get_tentacle_path, get_tentacle_value_path, \
    get_tentacle_nodes, get_tentacles_value_nodes, get_matrix_default_value_path, set_tentacle_value, \
    get_tentacle_value, get_tentacle_node, get_available_symbols, \
    is_tentacle_value_valid, is_tentacles_values_valid, get_evaluations_by_evaluator, get_available_time_frames
from octobot_evaluators.errors import UnsetTentacleEvaluation
from octobot_evaluators.matrices.matrices import Matrices


@pytest.mark.asyncio
async def test_set_tentacle_value():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    set_tentacle_value(matrix.matrix_id, tentacle_type="TA", tentacle_value=0, tentacle_path=["test-path"])
    assert matrix.get_node_at_path(["test-path"]).node_value == 0

    set_tentacle_value(matrix.matrix_id, tentacle_type="TA", tentacle_value="value", tentacle_path=
    get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA", exchange_name="binance"))
    assert matrix.get_node_at_path(
        get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA", exchange_name="binance")).node_value == "value"
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacle_value():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    assert not get_tentacle_value(matrix.matrix_id, tentacle_path=["Test-TA"])

    matrix.matrix.get_or_create_node(path=["Test-TA"])
    matrix.set_node_value(value_type="TA", value_path=["Test-TA"], value=0)
    assert get_tentacle_value(matrix.matrix_id, tentacle_path=["Test-TA"]) == 0

    matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA"))
    matrix.set_node_value(value_type="TA", value_path=get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA"),
                          value="test")
    assert get_tentacle_value(matrix.matrix_id, tentacle_path=get_tentacle_path(tentacle_type="TA",
                                                                                tentacle_name="Test-TA")) == "test"
    Matrices.instance().del_matrix(matrix.matrix_id)


def test_get_matrix_default_value_path():
    assert get_matrix_default_value_path(tentacle_type="TA", tentacle_name="Test-TA", exchange_name="binance") == \
           get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA", exchange_name="binance")
    assert get_matrix_default_value_path(tentacle_type="TA", tentacle_name="Test-TA",
                                         symbol="ETH", time_frame="1h") == \
           get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA") + \
           get_tentacle_value_path(symbol="ETH", time_frame="1h")
    assert get_matrix_default_value_path(tentacle_type="TA", tentacle_name="Test-TA",
                                         symbol="ETH", exchange_name="bitmex") == \
           get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA", exchange_name="bitmex") + \
           get_tentacle_value_path(symbol="ETH")


def test_get_tentacle_path():
    assert get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA", exchange_name="binance") == ["binance", "TA",
                                                                                                       "Test-TA"]
    assert get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA") == ["TA", "Test-TA"]
    assert get_tentacle_path(exchange_name="binance", tentacle_name="Test-TA") == ["binance", "Test-TA"]
    assert get_tentacle_path(tentacle_name="Test-TA") == ["Test-TA"]


def test_get_tentacle_value_path():
    assert get_tentacle_value_path() == []
    assert get_tentacle_value_path(symbol="BTC") == ["BTC"]
    assert get_tentacle_value_path(time_frame="1m") == ["1m"]
    assert get_tentacle_value_path(symbol="ETH", time_frame="1h") == ["ETH", "1h"]


@pytest.mark.asyncio
async def test_get_tentacle_nodes_on_root():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    created_node_1 = matrix.matrix.get_or_create_node(
        get_tentacle_path(tentacle_type="NO_TYPE", tentacle_name="Test-TA"))
    created_node_2 = matrix.matrix.get_or_create_node(
        get_tentacle_path(exchange_name="binance", tentacle_name="Test-TA-2"))
    created_node_3 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_name="Test-TA-3"))
    assert get_tentacle_nodes(matrix.matrix_id) == [matrix.get_node_at_path(get_tentacle_path(tentacle_type="NO_TYPE")),
                                                    matrix.get_node_at_path(get_tentacle_path(exchange_name="binance")),
                                                    created_node_3]

    assert get_tentacle_nodes(matrix.matrix_id, tentacle_type="TA") == []
    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="bitfinex") == []
    assert get_tentacle_nodes(matrix.matrix_id, tentacle_type="NO_TYPE") == [created_node_1]
    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="binance") == [created_node_2]
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacle_nodes_on_tentacle_type():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    created_node_1 = matrix.matrix.get_or_create_node(
        get_tentacle_path(tentacle_type="NO_TYPE", tentacle_name="Test-TA"))
    created_node_2 = matrix.matrix.get_or_create_node(
        get_tentacle_path(tentacle_type="TEST_TYPE", tentacle_name="Test-TA-2"))

    assert get_tentacle_nodes(matrix.matrix_id, tentacle_type="TA") == []
    assert get_tentacle_nodes(matrix.matrix_id, tentacle_type="NO_TYPE") == [created_node_1]
    assert get_tentacle_nodes(matrix.matrix_id, tentacle_type="TEST_TYPE") == [created_node_2]
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacle_nodes_on_exchange_name_and_tentacle_type():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    created_node_1 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="NO_TYPE",
                                                                        tentacle_name="Test-TA",
                                                                        exchange_name="binance"))
    created_node_2 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TEST_TYPE",
                                                                        tentacle_name="Test-TA-2",
                                                                        exchange_name="binance"))
    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="binance") == [
        matrix.get_node_at_path(get_tentacle_path(exchange_name="binance",
                                                  tentacle_type="NO_TYPE")),
        matrix.get_node_at_path(get_tentacle_path(exchange_name="binance",
                                                  tentacle_type="TEST_TYPE"))]
    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="binance", tentacle_type="NO_TYPE") == [created_node_1]
    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="binance", tentacle_type="TEST_TYPE") == [created_node_2]
    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="bitfinex") == []
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacle_nodes_on_exchange_name():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    created_node_1 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_name="Test-TA",
                                                                        exchange_name="binance"))
    created_node_2 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_name="Test-TA-2",
                                                                        exchange_name="binance"))

    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="bitfinex") == []
    assert get_tentacle_nodes(matrix.matrix_id, tentacle_type="NO_TYPE") == []
    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="binance") == [created_node_1, created_node_2]
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacle_nodes_on_multiple_tentacle_type():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    created_node_1 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA"))
    created_node_2 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA-2"))
    created_node_3 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA-3"))
    assert get_tentacle_nodes(matrix.matrix_id, tentacle_type="TA") == [created_node_1, created_node_2, created_node_3]
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacle_nodes_on_multiple_tentacle_type_and_exchange_name():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    created_node_1 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA",
                                                                        tentacle_name="Test-TA",
                                                                        exchange_name="binance"))
    created_node_2 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA",
                                                                        tentacle_name="Test-TA-2",
                                                                        exchange_name="binance"))
    created_node_3 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA",
                                                                        tentacle_name="Test-TA-3",
                                                                        exchange_name="binance"))
    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="binance", tentacle_type="TA") == [created_node_1,
                                                                                                 created_node_2,
                                                                                                 created_node_3]
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacle_nodes_mixed():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    created_node_1 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA",
                                                                        tentacle_name="Test-TA",
                                                                        exchange_name="binance"))
    created_node_2 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA",
                                                                        tentacle_name="Test-TA-2",
                                                                        exchange_name="bitfinex"))
    created_node_3 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA",
                                                                        tentacle_name="Test-TA-3",
                                                                        exchange_name="binance"))
    created_node_4 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TEST-TYPE",
                                                                        tentacle_name="Test-TA-4",
                                                                        exchange_name="binance"))
    created_node_5 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TEST-TYPE",
                                                                        tentacle_name="Test-TA-5",
                                                                        exchange_name="bitfinex"))
    created_node_6 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TEST-TYPE",
                                                                        tentacle_name="Test-TA-6"))
    created_node_7 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_name="Test-TA-7",
                                                                        exchange_name="bitfinex"))
    created_node_8 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_name="Test-TA-8"))
    created_node_9 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_name="Test-TA-9",
                                                                        exchange_name="binance"))
    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="binance", tentacle_type="TA") == [created_node_1,
                                                                                                 created_node_3]
    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="binance", tentacle_type="TEST-TYPE") == [created_node_4]

    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="bitfinex", tentacle_type="TEST-TYPE") == [created_node_5]

    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="bitfinex") == [
        matrix.get_node_at_path(get_tentacle_path(exchange_name="bitfinex", tentacle_type="TA")),
        matrix.get_node_at_path(get_tentacle_path(exchange_name="bitfinex", tentacle_type="TEST-TYPE")),
        created_node_7]

    assert get_tentacle_nodes(matrix.matrix_id, tentacle_type="TEST-TYPE") == [created_node_6]

    assert get_tentacle_nodes(matrix.matrix_id, exchange_name="binance") == [
        matrix.get_node_at_path(get_tentacle_path(exchange_name="binance", tentacle_type="TA")),
        matrix.get_node_at_path(get_tentacle_path(exchange_name="binance", tentacle_type="TEST-TYPE")),
        created_node_9]
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacles_value_nodes_with_symbol():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    created_node = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA"))
    created_node_2 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA-2"))
    btc_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="BTC"), starting_node=created_node)
    eth_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="ETH"), starting_node=created_node)
    btc_node_2 = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="BTC"), starting_node=created_node_2)
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_2],
                                     symbol="BTC") == [btc_node, btc_node_2]
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_2],
                                     symbol="ETH") == [eth_node]
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacles_value_nodes_with_time_frame():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    assert get_tentacle_value_path(time_frame="1m") == ["1m"]
    created_node = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA"))
    created_node_2 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA-2"))
    m_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(time_frame="1m"), starting_node=created_node)
    h_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(time_frame="1h"), starting_node=created_node)
    h_node_2 = matrix.matrix.get_or_create_node(get_tentacle_value_path(time_frame="1h"), starting_node=created_node_2)
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_2],
                                     time_frame="1h") == [h_node, h_node_2]
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_2],
                                     symbol="1m") == [m_node]
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacles_value_nodes_with_symbol_and_time_frame():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)
    assert get_tentacle_value_path(symbol="ETH", time_frame="1h") == ["ETH", "1h"]

    created_node = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA"))
    created_node_2 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA-2"))
    created_node_3 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA-3"))
    btc_h_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="BTC", time_frame="1h"),
                                                  starting_node=created_node)
    btc_m_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="BTC", time_frame="1m"),
                                                  starting_node=created_node_2)
    eth_h_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="ETH", time_frame="1h"),
                                                  starting_node=created_node_3)
    eth_m_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="ETH", time_frame="1m"),
                                                  starting_node=created_node_2)
    eth_d_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="ETH", time_frame="1d"),
                                                  starting_node=created_node)
    ltc_h_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="LTC", time_frame="1h"),
                                                  starting_node=created_node_2)
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_2, created_node_3],
                                     symbol="BTC", time_frame="1h") == [btc_h_node]
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_2, created_node_3],
                                     symbol="BTC") == [
               matrix.get_node_at_path(get_tentacle_value_path(symbol="BTC"), starting_node=created_node),
               matrix.get_node_at_path(get_tentacle_value_path(symbol="BTC"), starting_node=created_node_2)]
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_3],
                                     symbol="BTC") == [
               matrix.get_node_at_path(get_tentacle_value_path(symbol="BTC"), starting_node=created_node)]
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_3],
                                     symbol="BTC", time_frame="1m") == []
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node_3],
                                     symbol="BTC") == []
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_2, created_node_3],
                                     symbol="ETH") == [
               matrix.get_node_at_path(get_tentacle_value_path(symbol="ETH"), starting_node=created_node),
               matrix.get_node_at_path(get_tentacle_value_path(symbol="ETH"), starting_node=created_node_2),
               matrix.get_node_at_path(get_tentacle_value_path(symbol="ETH"), starting_node=created_node_3)]
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_tentacles_value_nodes_mixed():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)

    created_node = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA"))
    created_node_2 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA-2"))
    created_node_3 = matrix.matrix.get_or_create_node(get_tentacle_path(tentacle_type="TA", tentacle_name="Test-TA-3"))
    btc_h_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="BTC", time_frame="1h"),
                                                  starting_node=created_node)
    btc_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="BTC"), starting_node=created_node_2)
    eth_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="ETH"), starting_node=created_node_3)
    eth_m_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="ETH", time_frame="1m"),
                                                  starting_node=created_node_2)
    eth_d_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="ETH", time_frame="1d"),
                                                  starting_node=created_node)
    ltc_h_node = matrix.matrix.get_or_create_node(get_tentacle_value_path(symbol="LTC", time_frame="1h"),
                                                  starting_node=created_node_2)

    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_2, created_node_3],
                                     symbol="BTC", time_frame="1h") == [btc_h_node]
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_2, created_node_3],
                                     symbol="BTC") == [
               matrix.get_node_at_path(get_tentacle_value_path(symbol="BTC"), starting_node=created_node),
               matrix.get_node_at_path(get_tentacle_value_path(symbol="BTC"), starting_node=created_node_2)]
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_3],
                                     symbol="BTC") == [
               matrix.get_node_at_path(get_tentacle_value_path(symbol="BTC"), starting_node=created_node)]
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_3],
                                     symbol="BTC", time_frame="1m") == []
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node_3],
                                     symbol="BTC") == []
    assert get_tentacles_value_nodes(matrix.matrix_id, tentacle_nodes=[created_node, created_node_2, created_node_3],
                                     symbol="ETH") == [
               matrix.get_node_at_path(get_tentacle_value_path(symbol="ETH"), starting_node=created_node),
               matrix.get_node_at_path(get_tentacle_value_path(symbol="ETH"), starting_node=created_node_2),
               matrix.get_node_at_path(get_tentacle_value_path(symbol="ETH"), starting_node=created_node_3)]
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_is_tentacle_value_valid():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)

    evaluator_1_path = get_matrix_default_value_path(tentacle_type="TA", tentacle_name="Test-TA",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1m")
    evaluator_2_path = get_matrix_default_value_path(tentacle_type="TA", tentacle_name="Test-TA",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1h")

    # simulate AbstractEvaluator.initialize()
    set_tentacle_value(matrix.matrix_id, evaluator_1_path, "TA", None)
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None)

    get_tentacle_node(matrix.matrix_id, evaluator_1_path).node_value_time = time.time()
    assert is_tentacle_value_valid(matrix.matrix_id, evaluator_1_path)
    assert not is_tentacle_value_valid(matrix.matrix_id, evaluator_2_path)

    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None, timestamp=100)
    assert not is_tentacle_value_valid(matrix.matrix_id, evaluator_2_path)

    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None,
                       timestamp=time.time() - TimeFramesMinutes[TimeFrames.ONE_HOUR] * 2 * MINUTE_TO_SECONDS)
    assert not is_tentacle_value_valid(matrix.matrix_id, evaluator_2_path)

    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None,
                       timestamp=time.time() - TimeFramesMinutes[TimeFrames.ONE_HOUR] * MINUTE_TO_SECONDS)
    assert is_tentacle_value_valid(matrix.matrix_id, evaluator_2_path)

    # test delta
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None,
                       timestamp=time.time() - TimeFramesMinutes[TimeFrames.ONE_HOUR] * MINUTE_TO_SECONDS - 10)
    assert not is_tentacle_value_valid(matrix.matrix_id, evaluator_2_path)

    # test delta
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None,
                       timestamp=time.time() - TimeFramesMinutes[TimeFrames.ONE_HOUR] * MINUTE_TO_SECONDS - 9)
    assert is_tentacle_value_valid(matrix.matrix_id, evaluator_2_path)

    # test modified delta
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None,
                       timestamp=time.time() - TimeFramesMinutes[TimeFrames.ONE_HOUR] * MINUTE_TO_SECONDS - 29)
    assert is_tentacle_value_valid(matrix.matrix_id, evaluator_2_path, delta=30)

    # test modified delta
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None,
                       timestamp=time.time() - TimeFramesMinutes[TimeFrames.ONE_HOUR] * MINUTE_TO_SECONDS - 31)
    assert not is_tentacle_value_valid(matrix.matrix_id, evaluator_2_path, delta=30)

    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_is_tentacles_values_valid():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)

    evaluator_1_path = get_matrix_default_value_path(tentacle_type="TA", tentacle_name="Test-TA",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1m")
    evaluator_2_path = get_matrix_default_value_path(tentacle_type="TA", tentacle_name="Test-TA",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1h")

    # simulate AbstractEvaluator.initialize()
    set_tentacle_value(matrix.matrix_id, evaluator_1_path, "TA", None)
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None)

    assert not is_tentacles_values_valid(matrix.matrix_id, [evaluator_1_path, evaluator_2_path])

    set_tentacle_value(matrix.matrix_id, evaluator_1_path, "TA", None,
                       timestamp=time.time() - TimeFramesMinutes[TimeFrames.ONE_MINUTE] * 2 * MINUTE_TO_SECONDS)
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None,
                       timestamp=time.time() - TimeFramesMinutes[TimeFrames.ONE_HOUR] * 2 * MINUTE_TO_SECONDS)
    assert not is_tentacles_values_valid(matrix.matrix_id, [evaluator_1_path, evaluator_2_path])

    set_tentacle_value(matrix.matrix_id, evaluator_1_path, "TA", None,
                       timestamp=time.time() - TimeFramesMinutes[TimeFrames.ONE_MINUTE] * MINUTE_TO_SECONDS)
    assert not is_tentacles_values_valid(matrix.matrix_id, [evaluator_1_path, evaluator_2_path])

    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", None,
                       timestamp=time.time() - TimeFramesMinutes[TimeFrames.ONE_HOUR] * MINUTE_TO_SECONDS)
    assert is_tentacles_values_valid(matrix.matrix_id, [evaluator_1_path, evaluator_2_path])
    Matrices.instance().del_matrix(matrix.matrix_id)


@pytest.mark.asyncio
async def test_get_evaluations_by_evaluator():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)

    evaluator_1_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="RSI",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1m")
    evaluator_2_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="ADX",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1m")
    evaluator_3_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="ADX",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1h")

    # simulate AbstractEvaluator.initialize()
    set_tentacle_value(matrix.matrix_id, evaluator_1_path, "TA", 1)
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", -0.5)
    set_tentacle_value(matrix.matrix_id, evaluator_3_path, "TA", 0)

    assert get_evaluations_by_evaluator(matrix.matrix_id,
                                        tentacle_type="TA",
                                        exchange_name="kraken",
                                        cryptocurrency="BTC",
                                        symbol="BTC/USD",
                                        time_frame="1m") == {
               "RSI": get_tentacle_node(matrix.matrix_id, evaluator_1_path),
               "ADX": get_tentacle_node(matrix.matrix_id, evaluator_2_path)
           }

    # set invalid value to not to add this value in result dict
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", START_PENDING_EVAL_NOTE)
    with pytest.raises(UnsetTentacleEvaluation):
        get_evaluations_by_evaluator(matrix.matrix_id,
                                     tentacle_type="TA",
                                     exchange_name="kraken",
                                     cryptocurrency="BTC",
                                     symbol="BTC/USD",
                                     time_frame="1m",
                                     allow_missing=False)
    assert get_evaluations_by_evaluator(matrix.matrix_id,
                                        tentacle_type="TA",
                                        exchange_name="kraken",
                                        cryptocurrency="BTC",
                                        symbol="BTC/USD",
                                        time_frame="1m",
                                        allow_missing=True) == {
               "RSI": get_tentacle_node(matrix.matrix_id, evaluator_1_path)
           }
    assert get_evaluations_by_evaluator(matrix.matrix_id,
                                        tentacle_type="TA",
                                        exchange_name="kraken",
                                        cryptocurrency="BTC",
                                        symbol="BTC/USD",
                                        time_frame="1m",
                                        allow_missing=True,
                                        allowed_values=[START_PENDING_EVAL_NOTE]) == {
               "RSI": get_tentacle_node(matrix.matrix_id, evaluator_1_path),
               "ADX": get_tentacle_node(matrix.matrix_id, evaluator_2_path)
           }

    # invalid path
    assert get_evaluations_by_evaluator(matrix.matrix_id,
                                        tentacle_type="TA_invalid",
                                        exchange_name="kraken",
                                        cryptocurrency="BTC",
                                        symbol="BTC/USD",
                                        time_frame="1m",
                                        allow_missing=True,
                                        allowed_values=[START_PENDING_EVAL_NOTE]) == {}


@pytest.mark.asyncio
async def test_get_available_time_frames():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)

    evaluator_1_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="RSI",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1m")
    evaluator_2_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="ADX",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1m")
    evaluator_3_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="ADX",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1h")
    evaluator_4_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="RSI",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1h")

    # simulate AbstractEvaluator.initialize()
    set_tentacle_value(matrix.matrix_id, evaluator_1_path, "TA", 1)
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", -0.5)
    set_tentacle_value(matrix.matrix_id, evaluator_3_path, "TA", 0)
    set_tentacle_value(matrix.matrix_id, evaluator_4_path, "TA", -1)

    assert get_available_time_frames(matrix.matrix_id,
                                     exchange_name="kraken",
                                     tentacle_type="TA",
                                     cryptocurrency="BTC",
                                     symbol="BTC/USD") == ["1m", "1h"]

    # invalid path
    assert get_available_time_frames(matrix.matrix_id,
                                     exchange_name="kraken",
                                     tentacle_type="TA_invalid",
                                     cryptocurrency="BTC",
                                     symbol="BTC/USD") == []


@pytest.mark.asyncio
async def test_get_available_symbols():
    matrix = Matrix()
    Matrices.instance().add_matrix(matrix)

    evaluator_1_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="RSI",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1m")
    evaluator_2_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="RSI",
                                                     cryptocurrency="BTC",
                                                     symbol="BTCX/USDC",
                                                     time_frame="1m")
    evaluator_3_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="RSI",
                                                     cryptocurrency="BTC",
                                                     symbol="BTCX/USDC",
                                                     time_frame="1h")
    evaluator_4_path = get_matrix_default_value_path(tentacle_type="TA",
                                                     exchange_name="kraken",
                                                     tentacle_name="RSI",
                                                     cryptocurrency="BTC",
                                                     symbol="BTC/USD",
                                                     time_frame="1h")

    # simulate AbstractEvaluator.initialize()
    set_tentacle_value(matrix.matrix_id, evaluator_1_path, "TA", 1)
    set_tentacle_value(matrix.matrix_id, evaluator_2_path, "TA", -0.5)
    set_tentacle_value(matrix.matrix_id, evaluator_3_path, "TA", 0)
    set_tentacle_value(matrix.matrix_id, evaluator_4_path, "TA", -1)

    assert get_available_symbols(matrix.matrix_id,
                                 exchange_name="kraken",
                                 cryptocurrency="BTC") == ["BTC/USD", "BTCX/USDC"]

    # invalid path
    assert get_available_symbols(matrix.matrix_id, exchange_name="invalid_exchange", cryptocurrency="BTC") == []
    assert get_available_symbols(matrix.matrix_id, exchange_name="kraken", cryptocurrency="BTCX") == []
    assert get_available_symbols(matrix.matrix_id, exchange_name="invalid_exchange", cryptocurrency="BTCX") == []

    # now valid using real-time evaluation
    evaluator_5_path = get_matrix_default_value_path(tentacle_type="REAL_TIME",
                                                     exchange_name="kraken",
                                                     tentacle_name="RSI",
                                                     cryptocurrency="BTCX",
                                                     symbol="BTCX/USD",
                                                     time_frame="1h")
    set_tentacle_value(matrix.matrix_id, evaluator_5_path, "REAL_TIME", -1)
    assert get_available_symbols(matrix.matrix_id, exchange_name="kraken", cryptocurrency="BTCX") == ["BTCX/USD"]



# cython: language_level=3
#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from asyncio import CancelledError

from octobot_channels.constants import CHANNEL_WILDCARD
from octobot_commons.logging.logging_util import get_logger
from octobot_evaluators.channels.evaluator_channel import EvaluatorChannelConsumer, EvaluatorChannelProducer, \
    EvaluatorChannel
from octobot_evaluators.constants import EVALUATOR_EVAL_DEFAULT_TYPE
from octobot_evaluators.data_manager.matrix_manager import set_tentacle_value, get_matrix_default_value_path


class MatrixChannelConsumer(EvaluatorChannelConsumer):
    """
    EvaluatorChannelConsumer adapted for MatrixChannel
    """


class MatrixChannelProducer(EvaluatorChannelProducer):
    """
    EvaluatorChannelProducer adapted for MatrixChannel
    """

    # noinspection PyMethodOverriding
    async def send(self,
                   matrix_id,
                   evaluator_name,
                   evaluator_type,
                   eval_note,
                   eval_note_type=EVALUATOR_EVAL_DEFAULT_TYPE,
                   exchange_name=None,
                   cryptocurrency=CHANNEL_WILDCARD,
                   symbol=CHANNEL_WILDCARD,
                   time_frame=None,
                   origin_consumer=None):
        for consumer in self.channel.get_filtered_consumers(matrix_id=matrix_id,
                                                            cryptocurrency=cryptocurrency,
                                                            symbol=symbol,
                                                            time_frame=time_frame,
                                                            evaluator_type=evaluator_type,
                                                            evaluator_name=evaluator_name,
                                                            exchange_name=exchange_name,
                                                            origin_consumer=origin_consumer):
            await consumer.queue.put({
                "matrix_id": matrix_id,
                "evaluator_name": evaluator_name,
                "evaluator_type": evaluator_type,
                "eval_note": eval_note,
                "eval_note_type": eval_note_type,
                "exchange_name": exchange_name,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "time_frame": time_frame
            })

    async def send_eval_note(self,
                             matrix_id: str,
                             evaluator_name: str,
                             evaluator_type,
                             eval_note,
                             eval_note_type,
                             eval_time: float = 0,
                             exchange_name: str = None,
                             cryptocurrency: str = None,
                             symbol: str = None,
                             time_frame=None,
                             origin_consumer=None,
                             notify: bool = True):
        set_tentacle_value(
            matrix_id=matrix_id,
            tentacle_type=eval_note_type,
            tentacle_value=eval_note,
            timestamp=eval_time,
            tentacle_path=get_matrix_default_value_path(
                exchange_name=exchange_name,
                tentacle_type=evaluator_type,
                tentacle_name=evaluator_name,
                cryptocurrency=cryptocurrency,
                symbol=symbol,
                time_frame=time_frame
            )
        )
        if notify:
            await self.send(matrix_id=matrix_id,
                            evaluator_name=evaluator_name,
                            evaluator_type=evaluator_type,
                            eval_note=eval_note,
                            eval_note_type=eval_note_type,
                            exchange_name=exchange_name,
                            cryptocurrency=cryptocurrency,
                            symbol=symbol,
                            time_frame=time_frame,
                            origin_consumer=origin_consumer)


class MatrixChannel(EvaluatorChannel):
    FILTER_SIZE = 1
    PRODUCER_CLASS = MatrixChannelProducer
    CONSUMER_CLASS = MatrixChannelConsumer

    MATRIX_ID_KEY = "matrix_id"
    CRYPTOCURRENCY_KEY = "cryptocurrency"
    SYMBOL_KEY = "symbol"
    TIME_FRAME_KEY = "time_frame"
    EVALUATOR_TYPE_KEY = "evaluator_type"
    EXCHANGE_NAME_KEY = "exchange_name"
    EVALUATOR_NAME_KEY = "evaluator_name"

    def __init__(self, matrix_id):
        super().__init__(matrix_id)
        self.logger = get_logger(f"{self.__class__.__name__}")

    # noinspection PyMethodOverriding
    async def new_consumer(self,
                           callback: object,
                           size: int = 0,
                           priority_level: int = EvaluatorChannel.DEFAULT_PRIORITY_LEVEL,
                           matrix_id: str = CHANNEL_WILDCARD,
                           cryptocurrency: str = CHANNEL_WILDCARD,
                           symbol: str = CHANNEL_WILDCARD,
                           evaluator_name: str = CHANNEL_WILDCARD,
                           evaluator_type: object = CHANNEL_WILDCARD,
                           exchange_name: str = CHANNEL_WILDCARD,
                           time_frame=CHANNEL_WILDCARD) -> MatrixChannelConsumer:
        consumer = MatrixChannelConsumer(callback, size=size, priority_level=priority_level)
        await self._add_new_consumer_and_run(consumer,
                                             matrix_id=matrix_id,
                                             cryptocurrency=cryptocurrency,
                                             symbol=symbol,
                                             evaluator_name=evaluator_name,
                                             evaluator_type=evaluator_type,
                                             exchange_name=exchange_name,
                                             time_frame=time_frame)
        return consumer

    def get_filtered_consumers(self,
                               matrix_id=CHANNEL_WILDCARD,
                               cryptocurrency=CHANNEL_WILDCARD,
                               symbol=CHANNEL_WILDCARD,
                               evaluator_type=CHANNEL_WILDCARD,
                               time_frame=CHANNEL_WILDCARD,
                               evaluator_name=CHANNEL_WILDCARD,
                               exchange_name=CHANNEL_WILDCARD,
                               origin_consumer=None):
        return self.get_consumer_from_filters({
            self.MATRIX_ID_KEY: matrix_id,
            self.CRYPTOCURRENCY_KEY: cryptocurrency,
            self.SYMBOL_KEY: symbol,
            self.TIME_FRAME_KEY: time_frame,
            self.EVALUATOR_TYPE_KEY: evaluator_type,
            self.EVALUATOR_NAME_KEY: evaluator_name,
            self.EXCHANGE_NAME_KEY: exchange_name
        },
            origin_consumer=origin_consumer)

    async def _add_new_consumer_and_run(self, consumer,
                                        matrix_id=CHANNEL_WILDCARD,
                                        cryptocurrency=CHANNEL_WILDCARD,
                                        symbol=CHANNEL_WILDCARD,
                                        evaluator_name=CHANNEL_WILDCARD,
                                        evaluator_type=CHANNEL_WILDCARD,
                                        exchange_name=CHANNEL_WILDCARD,
                                        time_frame=None):
        consumer_filters: dict = {
            self.MATRIX_ID_KEY: matrix_id,
            self.CRYPTOCURRENCY_KEY: cryptocurrency,
            self.SYMBOL_KEY: symbol,
            self.TIME_FRAME_KEY: time_frame,
            self.EVALUATOR_NAME_KEY: evaluator_name,
            self.EXCHANGE_NAME_KEY: exchange_name,
            self.EVALUATOR_TYPE_KEY: evaluator_type
        }

        self.add_new_consumer(consumer, consumer_filters)
        await consumer.run()
        self.logger.debug(f"Consumer started for : "
                          f"[matrix_id={matrix_id},"
                          f" cryptocurrency={cryptocurrency},"
                          f" symbol={symbol},"
                          f" time_frame={time_frame},"
                          f" evaluator_name={evaluator_name},"
                          f" exchange_name={exchange_name},"
                          f" evaluator_type={evaluator_type}]")



# cython: language_level=3
#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_channels.channels.channel cimport Channel
from octobot_channels.consumer cimport Consumer, InternalConsumer, SupervisedConsumer
from octobot_channels.producer cimport Producer

cdef class EvaluatorChannel(Channel):
    cdef public str matrix_id

    cpdef list get_consumer_from_filters(self, dict consumer_filters, EvaluatorChannelConsumer origin_consumer=*)

cdef class EvaluatorChannelConsumer(Consumer):
    pass

cdef class EvaluatorChannelProducer(Producer):
    pass

cpdef EvaluatorChannel get_chan(str chan_name, str matrix_id)
cpdef dict get_evaluator_channels(str matrix_id)
cpdef void set_chan(EvaluatorChannel chan, str name)
cpdef void del_evaluator_channel_container(str matrix_id)
cpdef void del_chan(str name, str matrix_id)



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_commons.logging.logging_util import get_logger

from octobot_commons.singleton.singleton_class import Singleton
from octobot_evaluators.data.matrix import Matrix


class Matrices(Singleton):
    def __init__(self):
        self.matrices: dict = {}

    def add_matrix(self, matrix) -> None:
        if matrix.matrix_id not in self.matrices:
            self.matrices[matrix.matrix_id] = matrix

    def get_matrix(self, matrix_id) -> Matrix:
        return self.matrices[matrix_id]

    def del_matrix(self, matrix_id) -> None:
        try:
            if self.matrices[matrix_id]:
                self.matrices.pop(matrix_id, None)
        except KeyError:
            get_logger(self.__class__.__name__).warning(f"Can't del matrix with id {matrix_id}")



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_commons.enums import TimeFrames
from octobot_evaluators.evaluator.abstract_evaluator import AbstractEvaluator
from octobot_evaluators.util.evaluation_util import get_shortest_time_frame
from octobot_tentacles_manager.api.configurator import get_tentacle_config


class RealTimeEvaluator(AbstractEvaluator):
    __metaclass__ = AbstractEvaluator

    def __init__(self):
        super().__init__()
        self.load_config()

    def load_config(self):
        self.set_default_config()
        self.specific_config.update(get_tentacle_config(self.__class__))

    def get_symbol_candles(self, exchange_name: str, exchange_id: str, symbol: str, time_frame):
        try:
            from octobot_trading.api.symbol_data import get_symbol_candles_manager
            return get_symbol_candles_manager(self.get_exchange_symbol_data(exchange_name, exchange_id, symbol),
                                              time_frame)
        except ImportError:
            self.logger.error(f"Can't get candles manager: requires OctoBot-Trading package installed")

    def _get_tentacle_registration_topic(self, all_symbols_by_crypto_currencies, time_frames, real_time_time_frames):
        currencies, symbols, _ = super()._get_tentacle_registration_topic(all_symbols_by_crypto_currencies,
                                                                          time_frames,
                                                                          real_time_time_frames)
        to_handle_time_frames = []
        if self.time_frame is None:
            self.logger.error("Missing self.time_frame value, impossible to initialize this evaluator.")
        else:
            ideal_time_frame = TimeFrames(self.time_frame)
            to_handle_time_frame = get_shortest_time_frame(ideal_time_frame, real_time_time_frames, time_frames)
            if ideal_time_frame != to_handle_time_frame:
                self.logger.warning(f"Missing {ideal_time_frame.name} time frame in available time frames, "
                                    f"using {to_handle_time_frame.name} instead.")
            to_handle_time_frames = [to_handle_time_frame]
        # by default time frame registration only for the timeframe of this real-time evaluator
        return currencies, symbols, to_handle_time_frames

